import cerberus
from pathlib import Path
from typing import Union, List


class UniqueNameChecker:
    def __init__(self):
        self.name_lookup = []

    def check(self, field, value, error):
        if value in self.name_lookup:
            error(
                field,
                f'The build name must be unique. "{value}" has already been used.',
            )
        else:
            self.name_lookup.append(value)


class DefinesPrefixChecker:
    def check(self, field, defines: List[str], error):
        for value in defines:
            if value.startswith("-D"):
                error(
                    field,
                    f"Unnecessary -D prefix in {field}: {defines}. Aim will add this automatically.",
                )


class RequiresExistChecker:
    def __init__(self, document):
        self.doc = document

    def check(self, field, requires, error):
        builds = self.doc["builds"]
        for value in requires:
            for build in builds:
                if value == build["name"]:
                    break
            else:
                error(field, f"Build name does not exist: {value}")


class AbsProjectDirPathChecker:
    def check(self, field, paths, error):
        paths = [Path(the_path) for the_path in paths]

        for directory in paths:
            if not directory.is_absolute():
                error(field, f"Directory path should be absolute: {str(directory)}")
                break

            # Remember paths can now be directories or specific paths to files.
            if not directory.exists():
                error(field, f"Directory does not exist: {str(directory)}")
                break


class RelProjectDirPathChecker:
    def __init__(self, project_dir):
        self.project_dir = project_dir

    def check(self, field, paths, error):
        paths = [(self.project_dir / the_path) for the_path in paths]

        for directory in paths:
            # Remember paths can now be directories or specific paths to files.
            if not directory.exists():
                error(field, f'Directory does not exist: "{str(directory)}"')
                break


class AimCustomValidator(cerberus.Validator):
    def _check_with_output_naming_convention(self, field, value: Union[str, list]):
        # if you need more context then you can get it using the line below.
        # if self.document["buildRule"] in ["staticLib", "dynamicLib"]:

        # TODO: should we also check that the names are camelCase?
        # TODO: check outputNames are unique to prevent dependency cycle.

        def check_convention(_field, _value):
            the_errors = []
            if _value.startswith("lib"):
                the_error_str = f"Unnecessary 'lib' prefix in {_value}. Aim will add this automatically."
                the_errors.append(the_error_str)

            suffix = Path(_value).suffix
            if suffix:
                the_error_str = (
                    f'Unecessary suffix "{suffix}". Aim will add this automatically.'
                )
                the_errors.append(the_error_str)

            return the_errors

        # Bit of a hack so strings go through the same code path as lists.
        if isinstance(value, str):
            value = [value]

        for item in value:
            errors = check_convention(field, item)

            if errors:
                plural = ""
                if len(errors) > 1:
                    plural = "s"

                error_str = f"Naming convention error{plural}: {item}. " + " ".join(
                    errors
                )
                self._error(field, error_str)


def target_schema(document, project_dir):
    unique_name_checker = UniqueNameChecker()
    requires_exist_checker = RequiresExistChecker(document)
    path_checker = RelProjectDirPathChecker(project_dir)
    abs_path_checker = AbsProjectDirPathChecker()
    defines_checker = DefinesPrefixChecker()

    schema = {
        "compiler": {"required": True, "type": "string"},
        "archiver": {"required": True, "type": "string"},
        "compilerFrontend": {
            "required": True,
            "type": "string",
            "allowed": ["msvc", "gcc", "osx"],
        },
        "flags": {"type": "list", "schema": {"type": "string"}, "empty": False},
        "defines": {
            "type": "list",
            "schema": {"type": "string"},
            "empty": False,
            "check_with": defines_checker.check,
        },
        "projectRoot": {"required": True, "type": "string", "empty": False},
        "builds": {
            "required": True,
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "name": {
                        "required": True,
                        "type": "string",
                        "check_with": unique_name_checker.check,
                    },
                    "buildRule": {
                        "required": True,
                        "type": "string",
                        # "allowed": ["exe", "staticLib", "dynamicLib", "headerOnly", "libraryReference"],
                        "oneof": [
                            {
                                "excludes": "outputName",
                                "allowed": ["headerOnly", "libraryReference"],
                            },
                            {
                                "dependencies": ["outputName"],
                                "allowed": ["exe", "staticLib", "dynamicLib"],
                            },
                        ],
                    },
                    "compiler": {
                        "required": False,
                        "type": "string",
                        "dependencies": {
                            "buildRule": ["exe", "staticLib", "dynamicLib"]
                        },
                    },
                    "defines": {
                        "type": "list",
                        "schema": {"type": "string"},
                        "empty": False,
                        "check_with": defines_checker.check,
                        "dependencies": {
                            "buildRule": ["exe", "staticLib", "dynamicLib"]
                        },
                    },
                    "flags": {
                        "type": "list",
                        "schema": {"type": "string"},
                        "empty": False,
                        "dependencies": {
                            "buildRule": ["exe", "staticLib", "dynamicLib"]
                        },
                    },
                    "requires": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": requires_exist_checker.check,
                        "dependencies": {
                            "buildRule": ["exe", "staticLib", "dynamicLib"]
                        },
                    },
                    # Required but the requirement is handled by build rule.
                    "outputName": {
                        "type": "string",
                        "empty": False,
                        "check_with": "output_naming_convention",
                    },
                    "srcDirs": {
                        "required": False,
                        "empty": False,
                        "type": "list",
                        "schema": {"type": "string"},
                        "check_with": path_checker.check,
                        "dependencies": {
                            "buildRule": ["exe", "staticLib", "dynamicLib"]
                        },
                    },
                    "includePaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": path_checker.check,
                    },
                    "systemIncludePaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": abs_path_checker.check,
                        "dependencies": {"compilerFrontend": "gcc"},
                    },
                    "localIncludePaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": path_checker.check,
                        "dependencies": {"compilerFrontend": "gcc"},
                    },
                    "libraryPaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        # you can't check the library dirs as they may not exist if the project not built before.
                        # "check_with": path_checker.check,
                        "dependencies": {
                            "buildRule": ["exe", "dynamicLib", "libraryReference"]
                        },
                    },
                    "libraries": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": "output_naming_convention",
                        "dependencies": {
                            "buildRule": ["exe", "dynamicLib", "libraryReference"]
                        },
                    },
                },
            },
        },
    }

    validator = AimCustomValidator()
    validator.validate(document, schema)

    import pprint

    pretty = pprint.PrettyPrinter(indent=2, width=100)
    # TODO: Handle schema errors. https://docs.python-cerberus.org/en/stable/errors.html
    if validator.errors:
        for k, v in validator.errors.items():
            if k != "builds":
                print(f'Error for field "{k}"')
                pretty.pprint(f"{v}")
                print()

        builds_errors = validator.errors.get("builds", {})
        if builds_errors:
            assert len(builds_errors) == 1, "Build error list size is greater than 1."
            for k, v in builds_errors[0].items():
                builds = document["builds"]
                the_build = builds[k]
                the_build_name = the_build["name"]

                print(f'Error in build: "{the_build_name}"')
                assert (
                    len(v) == 1
                ), "Length is not 1. Not sure if it can ever be more than."
                pretty.pprint(v[0])
                print()
        exit(-1)



import functools
from typing import Dict, Tuple, Callable
from aim_build.utils import *
from ninja_syntax import Writer
from aim_build import commonbuilds
from pathlib import PurePosixPath

PrefixIncludePath = functools.partial(prefix, "-I")
PrefixSystemIncludePath = functools.partial(prefix, "-isystem")
PrefixQuoteIncludePath = functools.partial(prefix, "-iquote")
PrefixLibraryPath = functools.partial(prefix, "-L")
PrefixLibrary = functools.partial(prefix, "-l")
PrefixHashDefine = functools.partial(prefix, "-D")
ToObjectFiles = src_to_o


# TODO: Should take version strings as well?
def linux_add_static_library_naming_convention(library_name: str) -> str:
    return f"lib{library_name}.a"


def linux_add_dynamic_library_naming_convention(library_name: str) -> str:
    return f"lib{library_name}.so"


def linux_add_exe_naming_convention(exe_name: str) -> str:
    return f"{exe_name}.exe"


def get_quote_include_paths(build: Dict, build_dir: Path) -> List[PurePath]:
    include_paths = build.get("localIncludePaths", [])
    includes = relpaths(include_paths, build_dir)
    return includes


def get_system_include_paths(build: Dict) -> PathList:
    paths = build.get("systemIncludePaths", [])
    paths = [Path(path) for path in paths]
    return paths


def convert_strings_to_paths(paths):
    return [PurePosixPath(path) for path in paths]


def get_includes_for_build(build: Dict, parsed_toml: Dict) -> StringList:
    requires = [build["name"]] + build.get("requires", [])

    include_paths = set()
    system_include_paths = set()
    quote_include_paths = set()

    project_root = PurePosixPath(parsed_toml["projectRoot"])

    for required in requires:
        the_dep = commonbuilds.find_build(required, parsed_toml["builds"])

        includes = the_dep.get("includePaths", [])
        includes = convert_strings_to_paths(includes)
        includes = commonbuilds.get_include_paths(includes, project_root)
        include_paths.update(includes)

        quote_includes = get_quote_include_paths(the_dep, build["build_dir"])
        quote_include_paths.update(quote_includes)

        system_includes = get_system_include_paths(the_dep)
        system_include_paths.update(system_includes)

    include_paths = [str(path) for path in include_paths]
    system_include_paths = [str(path) for path in system_include_paths]
    quote_include_paths = [str(path) for path in quote_include_paths]

    include_paths.sort()
    system_include_paths.sort()
    quote_include_paths.sort()

    include_args = PrefixIncludePath(include_paths)
    system_include_args = PrefixSystemIncludePath(system_include_paths)
    quote_args = PrefixQuoteIncludePath(quote_include_paths)

    return include_args + system_include_args + quote_args


def get_external_libraries_paths(build: Dict) -> PathList:
    directory = build["directory"]
    library_paths = build.get("libraryPaths", [])
    library_paths = prepend_paths(directory, library_paths)
    return library_paths


def get_external_libraries_names(build: Dict) -> Tuple[StringList, StringList]:
    libraries = build.get("libraries", [])
    link_libraries = PrefixLibrary(libraries)
    return libraries, link_libraries


def get_external_libraries_information(build: Dict) -> Tuple[StringList, PathList]:
    libraries, _ = get_external_libraries_names(build)
    library_paths = get_external_libraries_paths(build)
    return libraries, library_paths


def get_src_for_build(build: Dict, parsed_toml: Dict) -> List[PurePosixPath]:
    files = commonbuilds.get_src_files(build, parsed_toml)
    return convert_strings_to_paths(files)


def add_compile_rule(
    writer: Writer,
    build: Dict,
    target_file: Dict,
    includes: StringList,
    extra_flags: StringList = None,
):
    build_name = build["name"]

    compiler, _, cxx_flags, defines = commonbuilds.get_toolchain_and_flags(
        build, target_file
    )
    defines = PrefixHashDefine(defines)

    if extra_flags:
        cxx_flags = extra_flags + cxx_flags

    src_files = get_src_for_build(build, target_file)
    obj_files = ToObjectFiles(src_files)
    obj_files = prepend_paths(Path(build_name), obj_files)

    file_pairs = zip(to_str(src_files), to_str(obj_files))
    for src_file, obj_file in file_pairs:
        writer.build(
            outputs=obj_file,
            rule="compile",
            inputs=src_file,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": cxx_flags,
                "defines": defines,
            },
        )
        writer.newline()

    return obj_files


def get_rpath(build: Dict, parsed_toml: Dict) -> str:
    # Good blog post about rpath:
    # https://medium.com/@nehckl0/creating-relocatable-linux-executables-by-setting-rpath-with-origin-45de573a2e98
    requires = build.get("requires", [])
    library_paths = set()

    # find_build_types("dynamicLib", parsed_toml["builds"])
    # TODO: replace the below with the above.
    for required in requires:
        the_dep = commonbuilds.find_build(required, parsed_toml["builds"])
        if the_dep["buildRule"] == "dynamicLib":
            library_paths.update([the_dep["name"]])

    top_build_dir = Path(build["build_dir"])
    build_dir = top_build_dir / build["name"]
    library_paths = list(library_paths)
    library_paths.sort()
    library_paths = prepend_paths(top_build_dir, library_paths)
    relative_paths = [relpath(Path(lib_path), build_dir) for lib_path in library_paths]

    relative_paths = [f"$$ORIGIN/{rel_path}" for rel_path in relative_paths]
    relative_paths = ["$$ORIGIN"] + relative_paths

    relative_paths_string = escape_path(":".join(relative_paths))
    return f"-Wl,-rpath='{relative_paths_string}'"


class GCCBuilds:
    def build(self, build: Dict, parsed_toml: Dict, ninja_writer: Writer, args):
        # TODO forward args
        build_name = build["name"]
        the_build = build["buildRule"]
        build_dir = build["build_dir"]

        build_path = build_dir / build_name
        build_path.mkdir(parents=True, exist_ok=True)

        build["buildPath"] = build_path

        if the_build == "staticLib":
            self.build_static_library(
                ninja_writer,
                build,
                parsed_toml,
                linux_add_static_library_naming_convention,
            )
        elif the_build == "exe":
            self.build_executable(ninja_writer, build, parsed_toml)
        elif the_build == "dynamicLib":
            self.build_dynamic_library(ninja_writer, build, parsed_toml)
        elif the_build == "headerOnly":
            pass
        elif the_build == "libraryReference":
            pass
        else:
            raise RuntimeError(f"Unknown build type {the_build}.")

    @staticmethod
    def build_static_library(
        pfw: Writer, build: Dict, parsed_toml: Dict, lib_name_func: Callable[[str], str]
    ):
        build_name = build["name"]

        includes = get_includes_for_build(build, parsed_toml)
        obj_files = add_compile_rule(pfw, build, parsed_toml, includes)

        library_name = lib_name_func(build["outputName"])
        relative_output_name = str(PurePosixPath(build_name) / library_name)

        _, archiver, cxx_flags, defines = commonbuilds.get_toolchain_and_flags(
            build, parsed_toml
        )
        defines = PrefixHashDefine(defines)

        pfw.build(
            outputs=relative_output_name,
            rule="archive",
            inputs=to_str(obj_files),
            variables={
                "archiver": archiver,
                "includes": includes,
                "flags": cxx_flags,
                "defines": defines,
            },
        )

        pfw.newline()
        pfw.build(rule="phony", inputs=relative_output_name, outputs=library_name)
        pfw.build(rule="phony", inputs=library_name, outputs=build_name)
        pfw.newline()

    @staticmethod
    def build_executable(pfw: Writer, build: Dict, parsed_toml: Dict):
        build_name = build["name"]

        compiler, _, cxxflags, defines = commonbuilds.get_toolchain_and_flags(
            build, parsed_toml
        )
        defines = PrefixHashDefine(defines)

        includes = get_includes_for_build(build, parsed_toml)
        obj_files = add_compile_rule(pfw, build, parsed_toml, includes)
        rpath = get_rpath(build, parsed_toml)
        (
            external_libraries_names,
            external_libraries_paths,
        ) = get_external_libraries_information(build)
        external_libraries_names = PrefixLibrary(external_libraries_names)
        external_libraries_paths = PrefixLibraryPath(external_libraries_paths)

        lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)
        requires_libraries = PrefixLibrary([info.name for info in lib_infos])
        requires_library_paths = PrefixLibraryPath([info.path for info in lib_infos])

        (
            ref_libraries,
            ref_library_paths,
        ) = commonbuilds.get_reference_library_information(build, parsed_toml)
        ref_libraries = PrefixLibrary(ref_libraries)
        ref_library_paths = PrefixLibraryPath(ref_library_paths)

        linker_args = (
            [rpath]
            + requires_library_paths
            + external_libraries_paths
            + ref_library_paths
            + requires_libraries
            + external_libraries_names
            + ref_libraries
        )

        full_library_names = commonbuilds.get_full_library_name_convention(
            lib_infos,
            linux_add_static_library_naming_convention,
            linux_add_dynamic_library_naming_convention,
        )

        exe_name = linux_add_exe_naming_convention(build["outputName"])
        relative_output_name = str(Path(build_name) / exe_name)
        pfw.build(
            rule="exe",
            outputs=relative_output_name,
            inputs=to_str(obj_files),
            implicit=full_library_names,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": cxxflags,
                "defines": defines,
                "linker_args": " ".join(linker_args),
            },
        )
        pfw.newline()
        pfw.build(rule="phony", inputs=relative_output_name, outputs=exe_name)
        pfw.build(rule="phony", inputs=exe_name, outputs=build_name)
        pfw.newline()

    @staticmethod
    def build_dynamic_library(pfw: Writer, build: Dict, parsed_toml: Dict):
        build_name = build["name"]

        extra_flags = ["-DEXPORT_DLL_PUBLIC", "-fvisibility=hidden", "-fPIC"]

        compiler, _, cxxflags, defines = commonbuilds.get_toolchain_and_flags(
            build, parsed_toml
        )
        defines = PrefixHashDefine(defines)

        includes = get_includes_for_build(build, parsed_toml)
        obj_files = add_compile_rule(pfw, build, parsed_toml, includes, extra_flags)
        rpath = get_rpath(build, parsed_toml)

        lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)
        requires_libraries = [info.name for info in lib_infos]
        requires_libraries = PrefixLibrary(requires_libraries)
        requires_library_paths = [info.path for info in lib_infos]
        requires_library_paths = PrefixLibraryPath(requires_library_paths)

        (
            external_libraries_names,
            external_libraries_paths,
        ) = get_external_libraries_information(build)
        external_libraries_names = PrefixLibrary(external_libraries_names)
        external_libraries_paths = PrefixLibraryPath(external_libraries_paths)

        (
            ref_libraries,
            ref_library_paths,
        ) = commonbuilds.get_reference_library_information(build, parsed_toml)
        ref_libraries = PrefixLibrary(ref_libraries)
        ref_library_paths = PrefixLibraryPath(ref_library_paths)

        linker_args = (
            [rpath]
            + requires_library_paths
            + external_libraries_paths
            + ref_library_paths
            + requires_libraries
            + external_libraries_names
            + ref_libraries
        )

        full_library_names = commonbuilds.get_full_library_name_convention(
            lib_infos,
            linux_add_static_library_naming_convention,
            linux_add_dynamic_library_naming_convention,
        )

        library_name = linux_add_dynamic_library_naming_convention(build["outputName"])
        relative_output_name = str(PurePosixPath(build_name) / library_name)
        pfw.build(
            rule="shared",
            inputs=to_str(obj_files),
            implicit=full_library_names,
            outputs=relative_output_name,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": " ".join(cxxflags),
                "defines": " ".join(defines),
                "lib_name": library_name,
                "linker_args": " ".join(linker_args),
            },
        )
        pfw.newline()
        pfw.build(rule="phony", inputs=relative_output_name, outputs=library_name)
        pfw.build(rule="phony", inputs=library_name, outputs=build_name)
        pfw.newline()


def add_naming_convention(
    output_name: str,
    build_type: str,
    static_convention_func=linux_add_static_library_naming_convention,
    dynamic_convention_func=linux_add_dynamic_library_naming_convention,
    exe_convention_func=linux_add_exe_naming_convention,
):
    if build_type == "staticLib":
        new_name = static_convention_func(output_name)
    elif build_type == "dynamicLib":
        new_name = dynamic_convention_func(output_name)
    else:
        new_name = exe_convention_func(output_name)

    return new_name


def log_build_information(build):
    build_name = build["name"]
    cxx_flags = build["global_flags"] + build.get("flags", [])
    defines = build["global_defines"] + build.get("defines", [])
    includes = build["includes"]
    library_paths = build["libraryPaths"]
    output = build["outputName"]

    print(f"Running build: f{build_name}")
    print(f"FLAGS: {cxx_flags}")
    print(f"DEFINES: {defines}")
    print(f"INCLUDE_PATHS: {includes}")
    print(f"LIBRARY_PATHS: {library_paths}")
    print(f"OUTPUT NAME: {output}")
    print("")



from typing import Dict, Tuple, Callable
from aim_build.utils import *
from dataclasses import dataclass

FileExtensions = ["*.cpp", "*.cxx", "*.cc", ".c"]


@dataclass
class LibraryInformation:
    name: str
    path: str
    type: str


def get_project_dir(build: Dict, target_file: Dict):
    root_dir = target_file["projectRoot"]
    project_dir = build["build_dir"] / root_dir
    return project_dir


def find_build(build_name: str, builds: Dict) -> Dict:
    # Note, this should never fail, as required dependencies are checked by the schema.
    for build in builds:
        if build["name"] == build_name:
            return build


def find_builds_of_type(build_type: str, builds: Dict) -> List[Dict]:
    return [build for build in builds if build["buildRule"] == build_type]


def get_include_paths(
    include_paths: List[PurePath], build_dir: PurePath
) -> List[PurePath]:
    abs_paths = [p for p in include_paths if p.is_absolute()]
    rel_paths = [p for p in include_paths if not p.is_absolute()]
    rel_paths = prepend_paths(build_dir, rel_paths)

    includes = abs_paths + rel_paths
    return includes


def get_toolchain_and_flags(
    build: Dict, target_file: Dict
) -> Tuple[str, str, StringList, StringList]:
    local_compiler = build.get("compiler", None)
    local_archiver = build.get("archiver", None)
    local_flags = build.get("flags", None)
    local_defines = build.get("defines", None)

    compiler = local_compiler if local_compiler else target_file["compiler"]
    archiver = local_archiver if local_archiver else target_file["archiver"]
    cxx_flags = local_flags if local_flags else target_file.get("flags", [])
    defines = local_defines if local_defines else target_file.get("defines", [])
    return compiler, archiver, cxx_flags, defines


def get_src_files(build: Dict, target_file: Dict) -> StringList:
    project_dir = get_project_dir(build, target_file)

    srcs = prepend_paths(project_dir, build["srcDirs"])
    src_dirs = [path for path in srcs if path.is_dir()]
    explicit_src_files = [path for path in srcs if path.is_file()]
    src_files = []
    for glob_pattern in FileExtensions:
        glob_files = flatten(glob(glob_pattern, src_dirs))
        src_files += glob_files

    src_files += explicit_src_files
    assert src_files, f"Fail to find any source files in {to_str(src_dirs)}."

    build_path = build["build_dir"]
    src_files = relpaths(src_files, build_path)

    return [str(file) for file in src_files]


def get_required_library_information(
    build: Dict, parsed_toml: Dict
) -> List[LibraryInformation]:
    requires = build.get("requires", [])
    if not requires:
        return []

    build_names = []  # Used to prevent duplicates.
    result = []

    for required in requires:
        the_dep = find_build(required, parsed_toml["builds"])
        if not the_dep["buildRule"] in ["staticLib", "dynamicLib"]:
            continue

        build_name = the_dep["name"]
        if build_name not in build_names:
            build_names.append(build_name)
            lib_info = LibraryInformation(
                the_dep["outputName"], the_dep["name"], the_dep["buildRule"]
            )
            result.append(lib_info)

    return result


def get_reference_library_information(
    build: Dict, parsed_toml: Dict
) -> Tuple[List[str], List[str]]:
    requires = build.get("requires", [])
    if not requires:
        return [], []

    build_names = []  # Used to prevent duplicates.
    libraries = []
    library_paths = []
    for required in requires:
        the_dep = find_build(required, parsed_toml["builds"])
        if the_dep["buildRule"] != "libraryReference":
            continue

        build_name = the_dep["name"]
        if build_name not in build_names:
            build_names.append(build_name)
            libraries += the_dep.get("libraries", [])
            library_paths += the_dep.get("libraryPaths", [])

    return libraries, library_paths


def get_full_library_name_convention(
    lib_infos: List[LibraryInformation],
    static_convention_func: Callable[[str], str],
    dynamic_convention_func: Callable[[str], str],
) -> StringList:
    # Here we just need to manage the fact that the linker's library flag (-l) needs the library name without
    # lib{name}.a/.so but the build dependency rule does need the full convention to find the build rule in the
    # build.ninja file.
    full_library_names = []
    for info in lib_infos:
        if info.type == "staticLib":
            full_library_names.append(static_convention_func(info.name))
        elif info.type == "dynamicLib":
            full_library_names.append(dynamic_convention_func(info.name))

    return full_library_names



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aim_build']

package_data = \
{'': ['*']}

install_requires = \
['cerberus>=1.3,<2.0',
 'ninja-syntax>=1.7,<2.0',
 'tabulate>=0.8.7,<0.9.0',
 'toml>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['aim = aim_build.main:entry']}

setup_kwargs = {
    'name': 'aim-build',
    'version': '0.1.30',
    'description': 'A powerful and easy to use build tool for C++.',
    'long_description': '<p align="center">\n<img src="https://github.com/diwalkerdev/Assets/blob/master/Aim/aim.png" width="300" height="300">\n</p>\n\n![GitHub release (latest SemVer including pre-releases)](https://img.shields.io/github/v/release/diwalkerdev/aim?include_prereleases)\n![GitHub commits since latest release (by SemVer including pre-releases)](https://img.shields.io/github/commits-since/diwalkerdev/aim/latest/dev?include_prereleases)\n![Python package](https://github.com/diwalkerdev/Aim/workflows/Python%20package/badge.svg?branch=dev)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/aim-build)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/aim-build)\n![GitHub contributors](https://img.shields.io/github/contributors/diwalkerdev/aim)\n![GitHub](https://img.shields.io/github/license/diwalkerdev/aim)\n\n\n# Aim\nA command line tool for building C++ projects. \n\n## Introduction\nAim is an attempt to make building C++ projects from source as simple as possible while encouraging a modular approach to software development.\n\nAim only requires a `target.toml` file which is used to specify the builds of your project. Each build specifies a\ncomponent of your project, like a static library, dynamic library, or an executable.\n\nAim supports:\n * Windows with `msvc` frontend.\n * Linux with `gcc` frontend.\n * It should also be possible to use the `gcc` frontend on Windows when using GCC-like compilers but this hasn\'t be tested.\n\n## Getting Started\n### Prerequisites\nAim requires the following dependencies:\n* [python](https://www.python.org/) - version 3.7 or above.\n* [ninja](https://ninja-build.org/)\n* [poetry](https://python-poetry.org/) - for development only\n\n### Installation\nAim is a `python` project and is installed using `pip`.\n\n```\npip install --user aim-build\n```\n\n### Using\n\nThere are 3 main commands:\n* `init` - initialises a directory with an empty project structure\n* `list --target=path/to/target_toml_dir` - displays the builds for the target\n* `build --target=path/to/target_toml_dir <build name>` - executes a build\n\nFor more information run:\n```\naim <command> --help\n```\n\nThe easiest way to get started is to use:\n\n`aim init --demo-files`\n\n`aim init` can be used to generate an empty\nproject structure and the `--demo-files` flags will copy a small test application into the current directory for \ndemonstration purposes.\n\nYou can then list the available builds of a target by specifying:\n\n`aim list --target=builds/linux-clang++-debug`\n\nAnd to build:\n\n`aim build --target=builds/linux-clang++-debug <build name>`\n\n<img src="https://github.com/diwalkerdev/Assets/blob/master/Aim/aim-init-demo.gif?raw=true" width="600px">\n\n## Target files\nA `target.toml` file describes a project and its build components.\n\nBegin by specifying `projectRoot` which is the path from the target file to your source files. All relative paths \nwill be relative to this directory.\n\nThe compiler frontend informs Aim how to construct the arguments for the compiler. Use `gcc`\nfor GCC-like compilers and `msvc` for Microsoft cl-like compilers. Next specify the `compiler`, `archiver`, `flags` and any `defines`. \n```\nprojectRoot = "../.."\n\ncompilerFrontend="gcc"\ncompiler = "clang++"\narchiver = "ar"\n\nflags = [\n    "-std=c++17",\n    "-O3",\n    "-g",\n    "-Wall",\n]\n\n# defines = [...] # Defines do not need the -D prefix.\n```\n\nNext specify your builds. For each build you must specify the `name` and `buildRule`. Valid build rules are `staticLib`,\n`dynamicLib`, `exe`, `headerOnly` or `libraryReference`. A build typically looks like:\n\n```\n[[builds]]\n    name = "calculatorApp"\n    buildRule = "exe"\n    requires = ["calculatorDynamic"] # A list of dependencies for this build.\n    outputName = "CalculatorApp"     # The output name. Aim will manage any prefixes or suffixes required.\n    srcDirs = ["src"]                # A list of source directories.\n    includePaths = ["include"]       # A list of include paths.\n    # The libraryPaths and libraries fields can be used to specify additional\n    # libraries and paths to the build. This allows for linking against third\n    # party libraries.\n    #libraryPaths = []\n    #libraries = []\n```\n\nOther notes:\n\n* The `requires` field is important as it is how you specify the dependencies for a build. For example, if you create a static library named "myAwesomeLibrary", this can be used in other builds simply by specifying  `requires=["myAwesomeLibrary"]`. \n\n* A `headerOnly` build does not have an `outputName` or `srcDirs` as it is not built. The `headerOnly` rule is not essential and is mostly for convenience. If you have a header only library, repeating the include paths across several builds can be become repetitive. Instead, create a `headerOnly` build to capture the include paths and use it in other builds by adding the rule to the builds `requires` field. \n\n* A `libraryReference` does not have `srcDirs` as it is not built. Like the `headerOnly` rule it is mostly for convience to reduce duplication. The primary use case is for capturing the `includePaths`, `libraryPaths` and `libraries` of a third party library that you need to use in a build. A `libraryReference` can then be used by other builds by adding it to a builds `requires` field.\n\n\n* The fields `compiler`, `flags` and `defines` are normally written at the top of the target file before the builds section. By default, all builds will use these fields i.e. they are global, but they can also be overridden by specifying them again in a build. Note that when these fields are specified specifically for a build, they completely replace the global definition; any `flags` or `defines` that you specify must be written out in full as they will not share any values with the global definition.\n\n## Supporting Multiple Targets\nAim treats any build variation as its own unique build target with its own unique `target.toml`. \n\nA build target is some combination of _things_ that affects the output binary such as:\n * operating system (Windows, OSX, Gnu Linux)\n * compiler (MSVC, GCC, Clang)\n * build type (Release, Debug, Sanitized)\n * etc. \n \nEach build target and corresponding `target.toml` file must have its own directory ideally named using a unique \nidentifier that comprises the \'parts\' that make up the build. For example, `builds/linux-clang++-release/target.toml` indicates that the target file describes a project that is a `release` build, uses the `clang++` compiler and is for the `linux` operating system. \n\nAs an example, if you were developing an application for both Windows and Linux, you may end up with a build directory structure like the following:\n * `builds/linux-clang++-release/target.toml`\n * `builds/linux-clang++-debug/target.toml`\n * `builds/windows-clangcl-release/target.toml`\n * `builds/windows-clangcl-debug/target.toml`\n\nNote: each `target.toml` file must be written out in full for each target that you need to support. There is no way for\ntarget files to share information or to depend on another. While this leads to duplication between target files, it \nmakes them very explicit and makes debugging builds much easier.\n\n## Advice Structuring Projects\nIf you structure your project/libraries as individual repositories then it may seem logical to nest dependencies inside \none another. For example, if library B depends on library A, then B needs a copy of A in order for it to be built.\nSo you may choose to nest the source of A inside B, perhaps using a git submodule.\n\nThe problem comes when your dependency hierarchy becomes more complex. If library C also depends on A, and an application D\ndepends on B and C, you\'ll end up with multiple copies of library A which can become difficult to manage.\n\nYou may need to use this approach, as it is useful to be able to build a library in isolation, however you should do so in such \na way where pulling the source for the dependencies is optional.\n\nThe approach the author uses is to use a non-project-specific directory that includes all your projects directly below it\ni.e. a "flat" structure. So rather than nesting dependencies you have:\n\n```\n + MyProjects\n + - LibA\n + - LibB\n + - LibC\n + - ApplicationD\n + - SomeOtherApp\n + - builds\n + - - linux-clang++-debug\n + - - - target.toml\n```\n\nThe flat structure has a single build directory and a single target file for each build target you need to support. This eliminates any \nduplication and is easy to manage. `Aim` is flexible enough that you can add additional levels to the project structure \nshould you need to. For example, you may want to group all libraries under a libraries sub-directory. But the take-away message \nis that you should not _enforce_ nested dependencies as this leads to duplication.\n\n\n## Developing Aim\n\nAim is a Python project and uses the [poetry](https://python-poetry.org/) dependency manager. See [poetry installation](https://python-poetry.org/docs/#installation) for instructions.\n\nOnce you have cloned the project, the virtual environment and dependencies can be installed by executing:\n\n```\npoetry install\n```\n\n### Dev Install\nUnfortunately, unlike `setuptools`, there is no means to do a \'dev install\' using poetry. A dev install effectively generates\nan application that internally references the active source files under development. This allows developers to test the application\nwithout having to re-install the application after each change.\n\nIn order to use a development version of Aim on the command line, is it recommended creating an alias. The alias needs to:\n* add the Aim directory to `PYTHONPATH` to resolve import/module paths \n* execute the main Aim script using the virtualenv created by poetry\n\nThere are `dev-env.bash` and `dev-env.fish` scripts that configure this for you in the root of the Aim project directory. \nNote, these files must be sourced in order for them to work. \n\n',
    'author': 'David Walker',
    'author_email': 'diwalkerdev@twitter.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/diwalkerdev/Aim',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import functools
from typing import Dict, Tuple
from ninja_syntax import Writer
from pathlib import PureWindowsPath, PurePosixPath
from typing import Callable
from aim_build.utils import *
from aim_build import commonbuilds

USING_RELATIVE_OUTPUTS = False

PrefixIncludePath = functools.partial(prefix, "/I")
PrefixLibraryPath = functools.partial(prefix, "/LIBPATH:")
PrefixLibrary = functools.partial(prefix, "")
PrefixHashDefine = functools.partial(prefix, "/D")
PostFixLib = functools.partial(postfix, ".lib")
ToObjectFiles = src_to_obj


# TODO: Should take version strings as well?
def windows_add_static_library_naming_convention(library_name: str) -> str:
    return f"lib{library_name}.lib"


def windows_add_dynamic_library_naming_convention(library_name: str) -> str:
    return f"lib{library_name}.dll"


def windows_add_exe_naming_convention(exe_name: str) -> str:
    return f"{exe_name}.exe"


def convert_posix_to_windows(paths: Union[StringList, List[PurePosixPath]]):
    return [str(path).replace("/", "\\") for path in paths]


def convert_strings_to_paths(paths: StringList):
    return [PureWindowsPath(path) for path in convert_posix_to_windows(paths)]


def convert_to_implicit_library_files(library):
    return "lib" + library + ".exp", "lib" + library + ".lib"


def implicit_library_outputs(libraries):
    implicits = []
    for library in libraries:
        parts = library.split(".")
        if parts[1] == "dll":
            implicits.append("lib" + parts[0] + ".exp")
            implicits.append("lib" + parts[0] + ".lib")

    return implicits


def convert_dlls_to_lib(libraries):
    new_libraries = []
    for library in libraries:
        parts = library.split(".")
        if parts[1] == "dll":
            new_libraries.append("lib" + parts[0] + ".lib")
        else:
            new_libraries.append("lib" + library)

    return new_libraries


def get_external_libraries_paths(build):
    directory = PureWindowsPath(build["directory"])
    library_paths = build.get("libraryPaths", [])
    library_paths = convert_posix_to_windows(library_paths)
    library_paths = prepend_paths(directory, library_paths)
    return library_paths


def get_external_libraries_names(build: Dict) -> Tuple[StringList, StringList]:
    libraries = build.get("libraries", [])
    link_libraries = PrefixLibrary(libraries)
    return libraries, link_libraries


def get_external_libraries_information(build: Dict) -> Tuple[StringList, PathList]:
    libraries, _ = get_external_libraries_names(build)
    library_paths = get_external_libraries_paths(build)
    return libraries, library_paths


def get_library_information(
    lib_infos: List[commonbuilds.LibraryInformation],
) -> Tuple[List, List, List, List]:
    exps = []
    libs = []
    # This is a bit confusing, we need just the library names and exp file names as this is what the linker needs, but
    # ninja needs the full path to the exp file.
    implicit_exps = []
    implicit_libs = []

    for info in lib_infos:
        if info.type == "dynamicLib":
            exp, lib = convert_to_implicit_library_files(info.name)
            exps.append(exp)
            libs.append(lib)
            if USING_RELATIVE_OUTPUTS:
                implicit_exps.append(str(PureWindowsPath(info.path) / exp))
                implicit_libs.append(str(PureWindowsPath(info.path) / lib))
            else:
                implicit_exps.append(exp)
                implicit_libs.append(lib)
        else:
            lib = "lib" + info.name + ".lib"
            libs.append(lib)

    return exps, libs, implicit_exps, implicit_libs


def get_includes_for_build(build: Dict, parsed_toml: Dict) -> StringList:
    requires = [build["name"]] + build.get("requires", [])

    include_paths = set()

    project_root = PureWindowsPath(parsed_toml["projectRoot"])

    for required in requires:
        the_dep = commonbuilds.find_build(required, parsed_toml["builds"])

        includes = the_dep.get("includePaths", [])
        includes = convert_strings_to_paths(includes)
        includes = commonbuilds.get_include_paths(includes, project_root)
        include_paths.update(includes)

    include_paths = [str(path) for path in include_paths]

    include_paths.sort()
    include_args = PrefixIncludePath(include_paths)

    return include_args


def get_src_for_build(build: Dict, parsed_toml: Dict) -> List[PureWindowsPath]:
    files = commonbuilds.get_src_files(build, parsed_toml)
    return convert_strings_to_paths(files)


def add_compile_rule(
    writer: Writer,
    build: Dict,
    target_file: Dict,
    includes: StringList,
    extra_flags: StringList = None,
):
    build_name = build["name"]

    compiler, _, cxx_flags, defines = commonbuilds.get_toolchain_and_flags(
        build, target_file
    )
    defines = PrefixHashDefine(defines)

    if extra_flags:
        cxx_flags = extra_flags + cxx_flags

    src_files = get_src_for_build(build, target_file)
    obj_files = ToObjectFiles(src_files)
    obj_files = prepend_paths(PureWindowsPath(build_name), obj_files)
    obj_files = convert_posix_to_windows(obj_files)

    file_pairs = zip(to_str(src_files), to_str(obj_files))
    for src_file, obj_file in file_pairs:
        writer.build(
            outputs=obj_file,
            rule="compile",
            inputs=src_file,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": cxx_flags,
                "defines": defines,
            },
        )
        writer.newline()

    return obj_files


class MSVCBuilds:
    def build(self, build: Dict, parsed_toml: Dict, ninja_writer: Writer, args):
        # TODO forward args
        build_name = build["name"]
        the_build = build["buildRule"]
        build_dir = build["build_dir"]

        build_path = build_dir / build_name
        build_path.mkdir(parents=True, exist_ok=True)

        build["buildPath"] = build_path

        if the_build == "staticLib":
            self.build_static_library(
                ninja_writer,
                build,
                parsed_toml,
                windows_add_static_library_naming_convention,
            )
        elif the_build == "exe":
            self.build_executable(ninja_writer, build, parsed_toml)
        elif the_build == "dynamicLib":
            self.build_dynamic_library(ninja_writer, build, parsed_toml)
        elif the_build == "headerOnly":
            pass
        elif the_build == "libraryReference":
            pass
        else:
            raise RuntimeError(f"Unknown build type {the_build}.")

    @staticmethod
    def build_static_library(
        pfw: Writer, build: Dict, parsed_toml: Dict, lib_name_func: Callable[[str], str]
    ):
        build_name = build["name"]

        includes = get_includes_for_build(build, parsed_toml)
        obj_files = add_compile_rule(pfw, build, parsed_toml, includes)

        library_name = lib_name_func(build["outputName"])
        relative_output_name = str(PureWindowsPath(build_name) / library_name)

        _, archiver, cxx_flags, defines = commonbuilds.get_toolchain_and_flags(
            build, parsed_toml
        )
        defines = PrefixHashDefine(defines)

        pfw.build(
            outputs=relative_output_name,
            rule="archive",
            inputs=to_str(obj_files),
            variables={
                "archiver": archiver,
                "includes": includes,
                "flags": cxx_flags,
                "defines": defines,
            },
        )

        pfw.newline()
        pfw.build(rule="phony", inputs=relative_output_name, outputs=library_name)
        pfw.build(rule="phony", inputs=library_name, outputs=build_name)
        pfw.newline()

    @staticmethod
    def build_executable(pfw: Writer, build: Dict, parsed_toml: Dict):
        build_name = build["name"]

        extra_flags = []

        compiler, _, cxxflags, defines = commonbuilds.get_toolchain_and_flags(
            build, parsed_toml
        )
        defines = PrefixHashDefine(defines)

        includes = get_includes_for_build(build, parsed_toml)
        obj_files = add_compile_rule(pfw, build, parsed_toml, includes, extra_flags)

        lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)

        (
            library_exps,
            link_libraries,
            implicit_exps,
            implicit_libs,
        ) = get_library_information(lib_infos)
        requires_libraries = PrefixLibrary(link_libraries)
        requires_library_paths = [info.path for info in lib_infos]
        requires_library_paths = PrefixLibraryPath(requires_library_paths)

        (
            external_libraries_names,
            external_libraries_paths,
        ) = get_external_libraries_information(build)
        external_libraries_names = PrefixLibrary(PostFixLib(external_libraries_names))
        external_libraries_paths = PrefixLibraryPath(external_libraries_paths)

        (
            ref_libraries,
            ref_library_paths,
        ) = commonbuilds.get_reference_library_information(build, parsed_toml)
        ref_libraries = PrefixLibrary(PostFixLib(ref_libraries))
        ref_library_paths = PrefixLibraryPath(
            convert_posix_to_windows(ref_library_paths)
        )

        linker_args = (
            requires_library_paths
            + external_libraries_paths
            + ref_library_paths
            + requires_libraries
            + external_libraries_names
            + ref_libraries
        )

        exe_name = windows_add_exe_naming_convention(build["outputName"])
        if USING_RELATIVE_OUTPUTS:
            build_output = str(PureWindowsPath(build_name) / exe_name)
        else:
            build_output = exe_name

        pfw.build(
            rule="exe",
            inputs=to_str(obj_files),
            implicit=implicit_exps + implicit_libs,
            outputs=build_output,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": " ".join(cxxflags),
                "defines": " ".join(defines),
                "exe_name": build_output,
                "linker_args": " ".join(linker_args),
            },
        )
        pfw.newline()
        if USING_RELATIVE_OUTPUTS:
            pfw.build(rule="phony", inputs=build_output, outputs=exe_name)
        pfw.build(rule="phony", inputs=exe_name, outputs=build_name)
        pfw.newline()

    @staticmethod
    def build_dynamic_library(pfw: Writer, build: Dict, parsed_toml: Dict):
        build_name = build["name"]

        extra_flags = ["/DEXPORT_DLL_PUBLIC"]

        compiler, _, cxxflags, defines = commonbuilds.get_toolchain_and_flags(
            build, parsed_toml
        )
        defines = PrefixHashDefine(defines)

        includes = get_includes_for_build(build, parsed_toml)
        obj_files = add_compile_rule(pfw, build, parsed_toml, includes, extra_flags)

        lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)

        (
            library_exps,
            link_libraries,
            implicit_exps,
            implicit_libs,
        ) = get_library_information(lib_infos)
        requires_libraries = PrefixLibrary(link_libraries)
        requires_library_paths = [info.path for info in lib_infos]
        requires_library_paths = PrefixLibraryPath(requires_library_paths)

        (
            external_libraries_names,
            external_libraries_paths,
        ) = get_external_libraries_information(build)
        external_libraries_names = PrefixLibrary(PostFixLib(external_libraries_names))
        external_libraries_paths = PrefixLibraryPath(external_libraries_paths)

        (
            ref_libraries,
            ref_library_paths,
        ) = commonbuilds.get_reference_library_information(build, parsed_toml)
        ref_libraries = PrefixLibrary(PostFixLib(ref_libraries))
        ref_library_paths = PrefixLibraryPath(
            convert_posix_to_windows(ref_library_paths)
        )

        linker_args = (
            requires_library_paths
            + external_libraries_paths
            + ref_library_paths
            + requires_libraries
            + external_libraries_names
            + ref_libraries
        )

        lib, exp = convert_to_implicit_library_files(build["outputName"])
        if USING_RELATIVE_OUTPUTS:
            implicit_outputs = [
                str(PureWindowsPath(build_name) / lib),
                str(PureWindowsPath(build_name) / exp),
            ]
        else:
            implicit_outputs = [lib, exp]

        library_name = windows_add_dynamic_library_naming_convention(
            build["outputName"]
        )

        if USING_RELATIVE_OUTPUTS:
            build_output = str(PureWindowsPath(build_name) / library_name)
        else:
            build_output = library_name

        pfw.build(
            rule="shared",
            inputs=to_str(obj_files),
            implicit=implicit_exps + implicit_libs,
            outputs=build_output,
            implicit_outputs=implicit_outputs,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": " ".join(cxxflags),
                "defines": " ".join(defines),
                "lib_name": build_output,
                "linker_args": " ".join(linker_args),
            },
        )
        pfw.newline()
        if USING_RELATIVE_OUTPUTS:
            pfw.build(rule="phony", inputs=build_output, outputs=library_name)
        pfw.build(rule="phony", inputs=library_name, outputs=build_name)
        pfw.newline()


def log_build_information(build):
    build_name = build["name"]
    cxxflags = build["flags"]
    defines = build["defines"]
    includes = build["includes"]
    library_paths = build["libraryPaths"]
    output = build["outputName"]

    print(f"Running build: f{build_name}")
    print(f"CXXFLAGS: {cxxflags}")
    print(f"DEFINES: {defines}")
    print(f"INCLUDE_PATHS: {includes}")
    print(f"LIBRARY_PATHS: {library_paths}")
    print(f"OUTPUT NAME: {output}")
    print("")
